@@ -128,11 +128,7 @@ impl Node {
 
 impl BlockData for Node {
     fn ext() -> &'static str {
-        super::EXT
-    }
-
-    fn max_size() -> u64 {
-        4096
+        "node"
     }
 }
 
@@ -246,7 +242,7 @@ where
 
     /// Create a new `BTreeFile`.
     pub async fn create(file: F, schema: RowSchema, txn_id: TxnId) -> TCResult<Self> {
-        if !file.is_empty(&txn_id).await? {
+        if !file.is_empty(txn_id).await? {
             return Err(TCError::internal(
                 "Tried to create a new BTree without a new File",
             ));
@@ -255,8 +251,9 @@ where
         let order = validate_schema(&schema)?;
 
         let root: BlockId = Uuid::new_v4().into();
+        let node = Node::new(true, None);
         file.clone()
-            .create_block(txn_id, root.clone(), Node::new(true, None))
+            .create_block(txn_id, root.clone(), node, DEFAULT_BLOCK_SIZE)
             .await?;
 
         Ok(BTreeFile::new(file, schema, order, root))
@@ -304,9 +301,6 @@ where
                     .map(|child_id| self._delete_range(txn_id, child_id, range))
                     .collect();
 
-                // free the write lock on this node before deleting from the child nodes
-                std::mem::drop(node);
-
                 deletes.try_fold((), |(), ()| future::ready(Ok(()))).await
             } else {
                 let child_id = node.children[r].clone();
@@ -315,12 +309,7 @@ where
         })
     }
 
-    fn _insert(
-        &self,
-        txn_id: TxnId,
-        node: <F::Block as Block<Node, F>>::ReadLock,
-        key: Key,
-    ) -> TCBoxTryFuture<()> {
+    fn _insert(&self, txn_id: TxnId, mut node: F::Write, key: Key) -> TCBoxTryFuture<()> {
         Box::pin(async move {
             let collator = &self.inner.collator;
             let file = &self.inner.file;
@@ -333,13 +322,15 @@ where
 
             if node.leaf {
                 let key = NodeKey::new(key);
-                let mut node = node.upgrade(file).await?;
 
                 if i == node.keys.len() {
                     node.keys.insert(i, key);
                     return Ok(());
                 }
 
+                #[cfg(debug_assertions)]
+                debug!("insert key {} into {} at {}", key, *node, i);
+
                 match collator.compare_slice(&key, &node.keys[i]) {
                     Ordering::Less => node.keys.insert(i, key),
                     Ordering::Equal => {
@@ -353,22 +344,16 @@ where
                 Ok(())
             } else {
                 let child_id = node.children[i].clone();
-                let child = file.read_block(txn_id, child_id).await?;
+                let child = file.write_block(txn_id, child_id).await?;
 
                 if child.keys.len() == (2 * order) - 1 {
-                    // split_child will need a write lock on child, so drop the read lock
-                    std::mem::drop(child);
-
                     let child_id = node.children[i].clone();
-                    let node = self
-                        .split_child(txn_id, child_id, node.upgrade(file).await?, i)
-                        .await?;
+                    let mut node = self.split_child(txn_id, node, child_id, child, i).await?;
 
                     match collator.compare_slice(&key, &node.keys[i]) {
                         Ordering::Less => self._insert(txn_id, node, key).await,
                         Ordering::Equal => {
                             if node.keys[i].deleted {
-                                let mut node = node.upgrade(file).await?;
                                 node.keys[i].deleted = false;
                             }
 
@@ -376,7 +361,8 @@ where
                         }
                         Ordering::Greater => {
                             let child_id = node.children[i + 1].clone();
-                            let child = file.read_block(txn_id, child_id).await?;
+
+                            let child = file.write_block(txn_id, child_id).await?;
                             self._insert(txn_id, child, key).await
                         }
                     }
@@ -514,7 +500,7 @@ where
     where
         Self: 'a,
     {
-        let root_id = self.inner.root.read(&txn_id).await?;
+        let root_id = self.inner.root.read(txn_id).await?;
 
         let root = self
             .inner
@@ -532,17 +518,17 @@ where
     async fn split_child(
         &self,
         txn_id: TxnId,
+        mut node: F::Write,
         node_id: NodeId,
-        mut node: <F::Block as Block<Node, F>>::WriteLock,
+        mut child: F::Write,
         i: usize,
-    ) -> TCResult<<F::Block as Block<Node, F>>::ReadLock> {
-        debug!("btree::split_child");
+    ) -> TCResult<F::Write> {
+        debug!("BTree::split_child");
 
         let file = &self.inner.file;
         let order = self.inner.order;
 
-        let child_id = node.children[i].clone(); // needed due to mutable borrow below
-        let mut child = file.write_block(txn_id, child_id.clone()).await?;
+        assert_eq!(node_id, node.children[i].clone());
 
         debug!(
             "child to split has {} keys and {} children",
@@ -550,12 +536,16 @@ where
             child.children.len()
         );
 
-        let new_node_id = file.unique_id(&txn_id).await?;
+        let new_node = Node::new(child.leaf, Some(node_id));
+        let (new_node_id, mut new_node) = file
+            .create_block_unique(txn_id, new_node, DEFAULT_BLOCK_SIZE)
+            .await?;
+
+        debug!("BTree::split_child created new node {}", new_node_id);
 
         node.children.insert(i + 1, new_node_id.clone());
         node.keys.insert(i, child.keys.remove(order - 1));
 
-        let mut new_node = Node::new(child.leaf, Some(node_id));
         new_node.keys = child.keys.drain((order - 1)..).collect();
 
         if child.leaf {
@@ -564,9 +554,7 @@ where
             new_node.children = child.children.drain(order..).collect();
         }
 
-        file.create_block(txn_id, new_node_id, new_node).await?;
-
-        node.downgrade(file).await
+        Ok(node)
     }
 }
 
@@ -602,7 +590,7 @@ where
     }
 
     async fn is_empty(&self, txn_id: TxnId) -> TCResult<bool> {
-        let root_id = self.inner.root.read(&txn_id).await?;
+        let root_id = self.inner.root.read(txn_id).await?;
         let root = self
             .inner
             .file
@@ -643,17 +631,17 @@ where
 
             self.inner.file.truncate(txn_id).await?;
 
-            *root = self.inner.file.unique_id(&txn_id).await?;
-
+            *root = Uuid::new_v4().into();
+            let node = Node::new(true, None);
             self.inner
                 .file
-                .create_block(txn_id, (*root).clone(), Node::new(true, None))
+                .create_block(txn_id, (*root).clone(), node, DEFAULT_BLOCK_SIZE)
                 .await?;
 
             return Ok(());
         }
 
-        let root_id = self.inner.root.read(&txn_id).await?;
+        let root_id = self.inner.root.read(txn_id).await?;
         self._delete_range(txn_id, (*root_id).clone(), &range).await
     }
 
@@ -668,7 +656,7 @@ where
         let mut root_id = self.inner.root.write(txn_id).await?;
         debug!("insert into BTree with root node ID {}", *root_id);
 
-        let root = file.read_block(txn_id, (*root_id).clone()).await?;
+        let root = file.write_block(txn_id, (*root_id).clone()).await?;
 
         #[cfg(debug_assertions)]
         debug!(
@@ -686,28 +674,25 @@ where
         assert_eq!(root.children.is_empty(), root.leaf);
 
         if root.keys.len() == (2 * order) - 1 {
-            // split_child will need a write lock on root, so release the read lock
-            std::mem::drop(root);
-
             debug!("split root node");
 
             let old_root_id = (*root_id).clone();
 
-            (*root_id) = file.unique_id(&txn_id).await?;
-
             let mut new_root = Node::new(false, None);
             new_root.children.push(old_root_id.clone());
 
-            let new_root = file
-                .create_block(txn_id, (*root_id).clone(), new_root)
+            let (new_root_id, new_root) = file
+                .create_block_unique(txn_id, new_root, DEFAULT_BLOCK_SIZE)
+                .await?;
+
+            (*root_id) = new_root_id;
+
+            let new_root = self
+                .split_child(txn_id, new_root, old_root_id, root, 0)
                 .await?;
 
-            let new_root = new_root.write().await;
-            let new_root = self.split_child(txn_id, old_root_id, new_root, 0).await?;
             self._insert(txn_id, new_root, key).await
         } else {
-            // no need to keep this write lock since we're not splitting the root node
-            std::mem::drop(root_id);
             self._insert(txn_id, root, key).await
         }
     }
@@ -741,12 +726,19 @@ impl<F: File<Node>, D: Dir, T: Transaction<D>> Persist<D> for BTreeFile<F, D, T>
     }
 
     async fn load(txn: &T, schema: RowSchema, file: F) -> TCResult<Self> {
+        debug!("BTreeFile::load {:?}", schema);
+
         let order = validate_schema(&schema)?;
 
         let txn_id = *txn.id();
         let mut root = None;
-        for block_id in file.block_ids(&txn_id).await? {
+        for block_id in file.block_ids(txn_id).await? {
+            debug!("BTreeFile::load block {}", block_id);
+
             let block = file.read_block(txn_id, block_id.clone()).await?;
+
+            debug!("BTreeFile::loaded block {}", block_id);
+
             if block.parent.is_none() {
                 root = Some(block_id);
                 break;
@@ -770,7 +762,7 @@ impl<F: File<Node>, D: Dir, T: Transaction<D>> Restore<D> for BTreeFile<F, D, T>
 
         let mut root_id = self.inner.root.write(txn_id).await?;
         self.inner.file.truncate(txn_id).await?;
-        *root_id = backup.inner.root.read(&txn_id).await?.clone();
+        *root_id = backup.inner.root.read(txn_id).await?.clone();
         self.inner.file.copy_from(&backup.inner.file, txn_id).await
     }
 }
